import logging
from typing import List, cast

import numpy as np

from cython_vst_loader.vst_event import VstMidiEvent, VstNoteOnMidiEvent, VstNoteOffMidiEvent
from cython_vst_loader.vst_host import VstHost
from cython_vst_loader.vst_plugin import VstPlugin as InternalPlugin
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.midi_event import MidiEvent
from py_headless_daw.schema.events.parameter_value_event import ParameterValueEvent
from py_headless_daw.schema.processing_strategy import ProcessingStrategy
from py_headless_daw.schema.unit import Unit


class VstPlugin(ProcessingStrategy):

    def __init__(self, path_to_plugin_file: bytes, unit: Unit):
        super().__init__()
        self.unit: Unit = unit
        self._path_to_plugin_file: bytes = path_to_plugin_file

        bpm = self.unit.host.bpm
        sample_rate = self.unit.host.sample_rate
        buffer_size = self.unit.host.block_size

        host = VstHost(int(sample_rate), buffer_size)
        host.bpm = bpm
        self._internal_plugin: InternalPlugin = InternalPlugin(path_to_plugin_file, host)

    def set_parameter_value(self, name: str, value: float):
        param_idx = self._find_parameter_index_by_name(name)
        self._internal_plugin.set_parameter_value(param_idx, value)

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        # noinspection PyTypeChecker
        midi_events: List[MidiEvent] = cast(List[MidiEvent], self._filter_events(event_inputs, Event.TYPE_MIDI))

        # noinspection PyTypeChecker
        parameter_events: List[ParameterValueEvent] = cast(List[ParameterValueEvent],
                                                           self._filter_events(event_inputs,
                                                                               Event.TYPE_PARAMETER_VALUE)
                                                           )

        # just setting the values of params (if any corresponding events have arrived) before
        # processing the upcoming audio block.

        # this is a simple, and, probably, wrong approach to setting parameters to a plugin:
        # it does not allow to precisely put parameter change event with sample
        # accuracy, also, the behaviour of plugin starts to depend on the block size.
        for parameter_event in parameter_events:
            parameter_index: int = self._find_parameter_index_by_name(parameter_event.parameter_id)
            self._internal_plugin.set_parameter_value(parameter_index, parameter_event.value)

        internal_midi_events = list(map(self._convert_midi_event_to_internal, midi_events))

        assert interval.num_samples > 0, "num samples in this interval <= 0: " + str(
            interval.num_samples) + ", this cannot be right"

        if len(midi_events) > 0:
            logging.debug('processing events for unit %s in interval %i' % (self.unit.name, interval.id))

            assert self.midi_events_are_valid(interval, midi_events)
            self._internal_plugin.process_events(internal_midi_events)

        def numpy_array_to_pointer(numpy_array: np.ndarray) -> int:
            if numpy_array.ndim != 1:
                raise Exception('expected a 1d numpy array here')
            pointer, ro_flag = numpy_array.__array_interface__['data']
            return pointer

        input_stream_pointers: List[int] = []
        output_stream_pointers: List[int] = []

        block_size: int = stream_outputs[0].shape[0]
        for stream_input in stream_inputs:
            if stream_input.shape[0] != block_size:
                raise Exception('block of different unexpected size')
            input_stream_pointers.append(numpy_array_to_pointer(stream_input))

        for stream_output in stream_outputs:
            if stream_output.shape[0] != block_size:
                raise Exception('block of different unexpected size')
            output_stream_pointers.append(numpy_array_to_pointer(stream_output))

        self._internal_plugin.process_replacing(input_stream_pointers, output_stream_pointers, block_size)

    @staticmethod
    def _convert_midi_event_to_internal(external_event: MidiEvent) -> VstMidiEvent:

        if not isinstance(external_event, MidiEvent):
            raise TypeError

        if external_event.is_note_on():
            res = VstNoteOnMidiEvent(
                external_event.get_sample_position(),
                external_event.note,
                external_event.velocity,
                external_event.channel
            )
        elif external_event.is_note_off():
            res = VstNoteOffMidiEvent(
                external_event.get_sample_position(),
                external_event.note,
                external_event.channel
            )
        else:
            raise TypeError('unknown external midi event type')

        return res

    def _find_parameter_index_by_name(self, parameter_string_id: str) -> int:
        """
        this method is used to convert from internal parameter ids (which are unique human-readable strings)
        into the indices of params of this particular vst plugin

        :param parameter_string_id:
        :return:
        """
        available_parameter_names: List[str] = []
        for idx in range(0, self._internal_plugin.get_num_parameters()):
            available_parameter_names.append(str(self._internal_plugin.get_parameter_name(idx)))
            if self._internal_plugin.get_parameter_name(idx) == parameter_string_id.encode('utf-8'):
                return idx

        raise LookupError(
            'a parameter named ' + str(parameter_string_id)
            + ' not found in this vst plugin. Available parameters are '
            + ", ".join(available_parameter_names))

    @staticmethod
    def midi_events_are_valid(interval: TimeInterval, midi_events: List[MidiEvent]) -> bool:
        for event in midi_events:
            if event.sample_position < 0:
                raise Exception('event has a negative sample position of ' + str(event.sample_position))
            if event.sample_position > interval.num_samples - 1:
                raise Exception(
                    'event has a too big sample position of ' + str(event.sample_position) + ". Max allowed is " + str(
                        interval.num_samples - 1)
                )

        return True



from typing import Dict, List, cast

from py_headless_daw.project.parameter import Parameter, ParameterValueType, ParameterRangeType


class HavingParameters:

    def __init__(self):
        self._parameters: Dict[str, Parameter] = {}
        super().__init__()

    def has_parameter(self, name: str) -> bool:
        return name in self._parameters

    def add_parameter(self,
                      name: str,
                      value: ParameterValueType,
                      param_type: str,
                      value_range: ParameterRangeType):

        if name in self._parameters:
            raise Exception('parameter named ' + name + ' already added to this object')

        parameter = Parameter(name, value, param_type, value_range)
        self._parameters[name] = parameter

    def add_parameter_object(self, parameter: Parameter) -> None:
        self._parameters[parameter.name] = parameter

    def get_parameter(self, name: str) -> Parameter:
        for parameter in self.parameters:
            if parameter.name == name:
                return parameter

        list_of_names: List[str] = [p.name for p in self.parameters]

        # noinspection PyTypeChecker
        available_names: List[str] = cast(List[str], list_of_names)

        raise Exception('parameter named ' + name + ' not found. Available: ' + ', '.join(available_names))

    def get_parameter_value(self, name: str) -> ParameterValueType:
        param = self.get_parameter(name)
        return param.value

    def get_float_parameter_value(self, name: str) -> float:
        param = self.get_parameter(name)
        if param.type != Parameter.TYPE_FLOAT:
            raise ValueError(f"parameter {name} was expected to be float (error: f009d0ef)")
        value = self.get_parameter_value(name)
        cast_value = cast(float, value)
        return cast_value

    def get_enum_parameter_value(self, name: str) -> str:
        param = self.get_parameter(name)
        if param.type != Parameter.TYPE_ENUM:
            raise ValueError(f"parameter {name} was expected to be enum (error: 80a1d180)")
        value = self.get_parameter_value(name)
        cast_value = cast(str, value)
        return cast_value

    def set_parameter_value(self, name: str, value: ParameterValueType):
        param = self.get_parameter(name)
        param.value = value

    @property
    def parameters(self) -> List[Parameter]:
        return list(self._parameters.values())



import numpy as np

from py_headless_daw.compiler.compiler_exception import CompilerException
from py_headless_daw.processing.hybrid.drum_synth_strategy import DrumSynthStrategy
from py_headless_daw.processing.stream.sampler import Sampler
from py_headless_daw.processing.stream.stereo_panner import StereoPanner
from py_headless_daw.processing.stream.stream_gain import StreamGain
from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.project.plugins.internal_plugin import GainPlugin, PanningPlugin, SamplerPlugin
from py_headless_daw.project.track import Track
from py_headless_daw.schema.processing_strategy import ProcessingStrategy
from py_headless_daw.services.wave_data_provider import WaveformProviderInterface
from py_headless_daw.project.plugins.internal_plugin import InternalPlugin as InternalProjectPlugin


class InternalPluginProcessingStrategyFactory:

    def __init__(self, waveform_provider: WaveformProviderInterface):
        self._waveform_provider: WaveformProviderInterface = waveform_provider

    # noinspection PyMethodMayBeStatic
    def produce(self, plugin: InternalProjectPlugin, track: Track) -> ProcessingStrategy:

        if isinstance(track, AudioTrack):
            overall_track_gain_value = track.get_gain_parameter().value
            overall_track_panning_value = track.get_panning_parameter().value

            if isinstance(plugin, GainPlugin):
                return StreamGain(np.float32(overall_track_gain_value))
            elif isinstance(plugin, PanningPlugin):
                return StereoPanner(np.float32(overall_track_panning_value))
            elif isinstance(plugin, SamplerPlugin):
                return Sampler(plugin.clips, self._waveform_provider)
            elif isinstance(plugin, DrumSynthPlugin):
                return self._construct_drum_synth_plugin(plugin)
            else:
                raise CompilerException(
                    "I don't know how to produce a processing strategy for internal plugin of type " + str(
                        type(plugin)))
        else:
            raise CompilerException(
                "I don't know how to compile internal plugins for this kind of tracks (error: 9029b61c)")

    def _construct_drum_synth_plugin(self, plugin: DrumSynthPlugin) -> DrumSynthStrategy:
        return DrumSynthStrategy(plugin)



from typing import List, cast

import numpy as np

from py_headless_daw.project.content.midi_clip import MidiClipEvent
from py_headless_daw.project.content.midi_note import MidiNote
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.midi_event import MidiEvent
from py_headless_daw.schema.events.midi_event_factory import MidiEventFactory
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class MidiTrackStrategy(ProcessingStrategy):
    def __init__(self, track: MidiTrack):
        super().__init__()
        self.track = track
        self._midi_event_factory: MidiEventFactory = MidiEventFactory()

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        clip_events: List[MidiClipEvent] = self.track.get_events(interval.start_in_seconds, interval.end_in_seconds)
        midi_event_lists: List[List[MidiEvent]] = []
        for event in clip_events:
            midi_event_lists.append(self._clip_event_transformer(interval, event))

        for event_output in event_outputs:
            for midi_event_list in midi_event_lists:
                for midi_event in midi_event_list:
                    event_output.append(midi_event)

    @staticmethod
    def round(a: float) -> float:
        return round(a, 5)

    def _clip_event_transformer(self, interval: TimeInterval, clip_event: MidiClipEvent) -> List[MidiEvent]:
        """
        the result can be multiple events

        :param interval:
        :param clip_event:
        :return:
        """
        if isinstance(clip_event, MidiNote):
            midi_note: MidiNote = cast(MidiNote, clip_event)
            return self._convert_midi_note_to_events(interval, midi_note)
        else:
            return []

    def _convert_midi_note_to_events(self, interval: TimeInterval, midi_note: MidiNote) -> List[MidiEvent]:
        note_start_in_seconds = midi_note.get_absolute_start_time()
        note_end_in_seconds = midi_note.get_absolute_end_time()

        if note_start_in_seconds > interval.end_in_seconds:
            raise Exception(
                'note start falls to the right of the interval: ' + str(note_start_in_seconds) + ' > ' + str(
                    interval.end_in_seconds) + '. It should not have happened, must be a bug')

        if note_end_in_seconds < interval.start_in_seconds:
            raise Exception('note ends falls to the left of the interval: ' + str(note_end_in_seconds) + ' < ' + str(
                interval.start_in_seconds) + '. It should not have happened, must be a bug')

        res: List[MidiEvent] = []

        if self.round(note_start_in_seconds) >= self.round(interval.start_in_seconds):
            sample_position = self._calculate_sample_position(interval, self.round(note_start_in_seconds))
            res.append(
                self._midi_event_factory.create_note_on_event(midi_note.note, midi_note.velocity, sample_position))

        if self.round(note_end_in_seconds) <= self.round(interval.end_in_seconds):
            sample_position = self._calculate_sample_position(interval, self.round(note_end_in_seconds))
            if sample_position == interval.num_samples:
                sample_position -= 1

            res.append(
                self._midi_event_factory.create_note_off_event(midi_note.note, midi_note.velocity, sample_position))

        return res

    @staticmethod
    def _calculate_sample_position(interval: TimeInterval, event_time_in_seconds: float) -> int:

        tolerance: float = 0.00000000001

        assert interval.start_in_seconds - tolerance <= event_time_in_seconds < interval.end_in_seconds + tolerance, \
            'event time %5.5f is out of interval [%5.5f, %5.5f)' % (
                event_time_in_seconds, interval.start_in_seconds, interval.end_in_seconds
            )

        relative_time_in_seconds = event_time_in_seconds - interval.start_in_seconds
        assert 0 - tolerance <= relative_time_in_seconds < interval.get_length_in_seconds() + tolerance

        return round(interval.num_samples * relative_time_in_seconds / interval.get_length_in_seconds())



import unittest

from py_headless_daw.project.envelope import Envelope, EnvelopePoint
from py_headless_daw.project.parameter import Parameter


class EnvelopeTest(unittest.TestCase):
    def test_envelope(self):
        target = Parameter('a', 0.777, Parameter.TYPE_FLOAT, (0.0, 1.0))
        envelope = Envelope(target)

        point1 = EnvelopePoint(0.0, 0.0)
        point2 = EnvelopePoint(1.0, 10.0)

        envelope.points = [point2, point1]

        self.assertEqual(0.0, envelope.get_value(0.0))
        self.assertEqual(5, envelope.get_value(0.5))
        self.assertEqual(10, envelope.get_value(1.0))
        self.assertEqual(0, envelope.get_value(-1.0))
        self.assertEqual(10, envelope.get_value(2.0))



import unittest

import parametrized as parametrized

from py_headless_daw.project.content.clip import Clip
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.shared.clip_intersection import ClipIntersection


@parametrized.parametrized_test_case
class ClipIntersectionTest(unittest.TestCase):

    @parametrized.parametrized([
        (1, 4, 2, 3, 2, 3, 0, 1),
        (1, 2, 3, 4, None),
        (3, 4, 1, 2, None),
        (1, 3, 2, 4, 2, 3, 0, 1)
    ])
    def get_intersection_with_interval(self,
                                       interval_start: float,
                                       interval_end: float,
                                       clip_start: float,
                                       clip_end: float,
                                       expected_intersection_start: float = None,
                                       expected_intersection_end: float = None,
                                       expected_intersection_start_clip_time: float = None,
                                       expected_intersection_end_clip_time: float = None):
        interval = TimeInterval()
        interval.start_in_seconds = interval_start
        interval.end_in_seconds = interval_end

        clip = Clip(clip_start, clip_end)

        intersection = ClipIntersection.create_intersection_of_clip_and_interval(clip, interval)

        if expected_intersection_start is None:
            self.assertIsNone(intersection)
            return

        self.assertEqual(expected_intersection_start, intersection.start_project_time,
                         "unexpected intersection start project time")
        self.assertEqual(expected_intersection_end, intersection.end_project_time,
                         "unexpected intersection end project time")
        self.assertEqual(expected_intersection_start_clip_time, intersection.start_clip_time,
                         "unexpected intersection start clip time")
        self.assertEqual(expected_intersection_end_clip_time, intersection.end_clip_time,
                         "unexpected intersection end clip time")

        self.assertIs(clip, intersection.clip)


if __name__ == '__main__':
    unittest.main()



import logging

from py_headless_daw.application_container import ApplicationContainer
from py_headless_daw.demo.riding_on_bugs.riding_on_bugs import RidingOnBugs
from pathlib import Path


def set_up_logging(level):
    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    root_logger.addHandler(handler)

    root_logger.setLevel(level)
    handler.setLevel(level)


if __name__ == '__main__':

    set_up_logging(logging.DEBUG)

    container = ApplicationContainer()

    demo = RidingOnBugs(container.project_renderer())
    out_file = str(Path(__file__).parent.absolute()) + "/output/out.wav"
    demo.render(out_file)



import os
from typing import List, Dict, Iterator

from py_headless_daw.project.parameter import Parameter
from py_headless_daw.project.named_parameter_bag import NamedParameterBag


class AmsynthPatchesManager:
    def __init__(self, patch_dir: str):
        """
        creates a manager object and reads the data from the supplied dir

        patches are organized into groups (each group is produced
        by a file found in a patch_dir)
        """
        self._patches: Dict[str, List[NamedParameterBag]] = {}
        self._read_files(patch_dir)

    def get_all_patches(self) -> List[NamedParameterBag]:
        return list(self._get_all_patches_iterator())

    def _get_all_patches_iterator(self) -> Iterator[NamedParameterBag]:
        for group in self._patches:
            for patch in self._patches[group]:
                yield patch

    def get_all_patches_from_group(self, group_name: str) -> List[NamedParameterBag]:
        if group_name not in self._patches:
            raise ValueError(
                f'group named {group_name} not found in this amsynth patches manager. Available groups: {" ".join(self._patches.keys())} (error: 3fee472e)')

        return self._patches[group_name]

    def _read_files(self, patch_dir: str):
        patch_files = os.listdir(patch_dir)
        for file in patch_files:
            self._patches[file] = self._read_one_patch_file(patch_dir, file)

    def _read_one_patch_file(self, patch_dir: str, file: str) -> List[NamedParameterBag]:
        with open(patch_dir + '/' + file) as f:
            lines = [x.strip() for x in f.readlines()]

        if "amSynth" != lines[0]:
            raise ValueError(f'file named {file} in {patch_dir} does not have an expected header (error: b6f8d217)')

        patch_lines_accumulator: List[str] = []
        patches: List[NamedParameterBag] = []
        for line in lines:

            if "amSynth" == line:
                continue

            if line.startswith('<preset> <name>'):
                if patch_lines_accumulator:
                    new_patch = self._produce_one_patch(patch_lines_accumulator)
                    patches.append(new_patch)
                    patch_lines_accumulator = []
            patch_lines_accumulator.append(line)
        return patches

    def _produce_one_patch(self, patch_strings_accumulator: List[str]) -> NamedParameterBag:
        patch = NamedParameterBag()

        for s in patch_strings_accumulator:
            if s.startswith("<preset> <name>"):
                patch.name = s.replace("<preset> <name> ", "").strip()
            elif s.startswith('<parameter>'):
                param_name: str = s.split(sep=" ")[1]
                param_value: float = float(s.split(sep=" ")[2])
                patch.add_parameter(
                    param_name,
                    param_value,
                    Parameter.TYPE_FLOAT,
                    value_range=(-100.0, 100.0)
                )
        return patch



import unittest

import numpy as np

from py_headless_daw.processing.stream.impulse_noice_generator import ImpulseNoiseGeneratorStrategy
from py_headless_daw.schema.dto.time_interval import TimeInterval


class ImpulseNoiseGeneratorStrategyTest(unittest.TestCase):

    @staticmethod
    def test_impulse_noise_generator_strategy():
        strategy = ImpulseNoiseGeneratorStrategy()
        out_stream_buffer = np.zeros(shape=(100,), dtype=np.float32)

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1

        strategy.render(interval, [], [out_stream_buffer], [], [])



from typing import List, cast, Dict

import numpy as np

from py_headless_daw.dto.waveform import Waveform
from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.schema.clip_track_processing_strategy import ClipTrackProcessingStrategy
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event

from py_headless_daw.services.wave_data_provider import WaveformProviderInterface
from py_headless_daw.shared.clip_intersection import ClipIntersection


class ProcessedWavData(Waveform):
    pass


class Sampler(ClipTrackProcessingStrategy):
    """
    Sampler is a strategy that produces audio basing on the number of associated AudioClips
    """

    _processed_wav_data_cache: Dict[str, ProcessedWavData] = {}

    def __init__(self, clips: List[AudioClip], wave_data_provider: WaveformProviderInterface):
        super().__init__(clips)
        for clip in clips:
            if not isinstance(clip, AudioClip):
                raise Exception(f"a clips was expected to be an AudioClip, instead it is {type(clip)}")
        self.clips: List[AudioClip] = clips
        self.wave_data_provider: WaveformProviderInterface = wave_data_provider

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        if self.unit is None:
            raise Exception('unit is not set in this sampler')

        [o.fill(0.0) for o in stream_outputs]
        intersections = self._find_intersections(interval)

        for intersection in intersections:
            self._render_one_intersection(intersection, stream_outputs, interval, self.unit.host.sample_rate)

    def _render_one_intersection(self, intersection: ClipIntersection, stream_outputs: List[np.ndarray],
                                 interval: TimeInterval,
                                 sample_rate: int):

        clip: AudioClip = cast(AudioClip, intersection.clip)
        wav_data = self._get_processed_wav_data(clip)

        if wav_data.num_channels != 1:
            # number of channels should match it not a mono sample
            if wav_data.num_channels != len(stream_outputs):
                raise Exception(
                    f"""number of channels in wav_data {wav_data.num_channels}
                    and in output {len(stream_outputs)} does not match. Related file: {clip.source_file_path}""")

        if intersection.start_clip_time is None:
            raise Exception('intersection is missing start clip time prop')

        patch_start_in_wav_data_in_samples: int = round(clip.cue_sample + intersection.start_clip_time * sample_rate)

        if patch_start_in_wav_data_in_samples > wav_data.length_in_samples():
            return

        if intersection.end_clip_time is None:
            raise Exception('intersection is missing end clip time prop')

        patch_end_in_wav_data_in_samples: int = min(
            round(clip.cue_sample + intersection.end_clip_time * sample_rate),
            wav_data.length_in_samples())

        if intersection.start_project_time is None:
            raise Exception('intersection is missing start project time')

        patch_start_in_output_in_samples: int = round(
            (intersection.start_project_time - interval.start_in_seconds) * sample_rate)

        patch_length_in_samples: int = patch_end_in_wav_data_in_samples - patch_start_in_wav_data_in_samples
        patch_end_in_output_in_samples: int = patch_start_in_output_in_samples + patch_length_in_samples

        patch_length_in_wav_data = patch_end_in_wav_data_in_samples - patch_start_in_wav_data_in_samples
        patch_length_in_output = patch_end_in_output_in_samples - patch_start_in_output_in_samples

        if patch_length_in_wav_data != patch_length_in_output:
            raise Exception(
                f"""patch lengths in sample and in interval diff. In sample: {patch_length_in_wav_data},
                in interval: {patch_length_in_output}""")

        for i, output in enumerate(stream_outputs):
            if wav_data.num_channels > 1:
                channel_in_wav = wav_data.data[i]
            else:
                # mono samples is a special case - we add their single channel uniformly to all outputs
                channel_in_wav = wav_data.data[0]

            # slicing returns a view, so it must be efficient
            patch_data: np.ndarray = channel_in_wav[patch_start_in_wav_data_in_samples:patch_end_in_wav_data_in_samples]
            patched_data: np.ndarray = output[patch_start_in_output_in_samples:patch_end_in_output_in_samples]

            if patch_data.shape != patched_data.shape:
                raise Exception(f"""patch and patched have different shapes.
                patch: {str(patch_data.shape)}
                patched: {str(patched_data.shape)}
                """)

            np.add(patch_data, patched_data, out=patched_data)

    def _get_processed_wav_data(self, clip: AudioClip) -> ProcessedWavData:
        cache_key: str = self._generate_cache_key(clip)
        if cache_key not in self._processed_wav_data_cache:
            self._processed_wav_data_cache[cache_key] = self._generate_processed_wav_data(clip)

        return self._processed_wav_data_cache[cache_key]

    def _generate_cache_key(self, clip: AudioClip) -> str:
        return clip.source_file_path + "_" + str(clip.rate)

    def _generate_processed_wav_data(self, clip: AudioClip) -> ProcessedWavData:

        raw_waveform = self.wave_data_provider.get_wave_data_by_file_path(clip.source_file_path)
        (sample_rate, data) = (raw_waveform.sample_rate, raw_waveform.data)

        # in the future, this will involve not just reading the wav file, but also
        # converting it into host's sample rate (self.unit.hot.sample_rate),
        # and taking into account clip's rate property.
        # Also reading non-wav files.

        # that's why Sample does not have sample rate property as it is already expected
        # to be in host's one

        return ProcessedWavData(
            data=data,
            sample_rate=sample_rate
        )



from typing import List

from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.project.plugins.internal_plugin import SamplerPlugin
from py_headless_daw.project.plugins.plugin import Plugin


class SamplerTrack(AudioTrack):
    """
    A sampler track is just an audio track that has some associated AudioClips
    When compiling, this it is also rendered similarly except one additional
    "Sampler" plugin added in the beginning of the chain.
    This initial plugin just emits audio according to the clip.
    """

    def __init__(self):
        super().__init__()
        self.clips: List[AudioClip] = []

    @AudioTrack.plugins.getter  # type: ignore
    def plugins(self) -> List[Plugin]:
        sampler_plugin: Plugin = SamplerPlugin(self.clips)
        return [sampler_plugin] + super().plugins



from typing import List

import numpy as np

from py_headless_daw.dsp_utils.adsr_envelope import ADSREnvelope
from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator_config import DrumSynthGeneratorConfig
from py_headless_daw.dsp_utils.drum_synth.one_shot_oscillator import OneShotOscillator
from py_headless_daw.dsp_utils.wave_producer_interface import WaveProducerInterface


class DrumSynthGenerator(WaveProducerInterface):

    def __init__(self, config: DrumSynthGeneratorConfig):
        self._oscillators: List[OneShotOscillator] = []
        for oscillator_config in config.oscillator_configs:
            new_osc = OneShotOscillator()
            volume_envelope = ADSREnvelope(oscillator_config.volume_envelope_attack_time,
                                           oscillator_config.volume_envelope_decay_time,
                                           oscillator_config.volume_envelope_sustain_level,
                                           oscillator_config.volume_envelope_sustain_time,
                                           oscillator_config.volume_envelope_release_time)

            volume_envelope.attack_curve_power = oscillator_config.volume_envelope_attack_curve_power
            volume_envelope.attack_curve = oscillator_config.volume_envelope_attack_curve_ratio

            volume_envelope.decay_curve_power = oscillator_config.volume_envelope_decay_curve_power
            volume_envelope.decay_curve = oscillator_config.volume_envelope_decay_curve_ratio

            volume_envelope.release_curve_power = oscillator_config.volume_envelope_release_curve_power
            volume_envelope.release_curve = oscillator_config.volume_envelope_release_curve_ratio

            pitch_envelope = ADSREnvelope(oscillator_config.pitch_envelope_attack_time,
                                          oscillator_config.pitch_envelope_decay_time,
                                          oscillator_config.pitch_envelope_sustain_level,
                                          oscillator_config.pitch_envelope_sustain_time,
                                          oscillator_config.pitch_envelope_release_time)

            pitch_envelope.attack_curve_power = oscillator_config.volume_envelope_attack_curve_power
            pitch_envelope.attack_curve = oscillator_config.volume_envelope_attack_curve_ratio

            pitch_envelope.decay_curve_power = oscillator_config.pitch_envelope_decay_curve_power
            pitch_envelope.decay_curve = oscillator_config.pitch_envelope_decay_curve_ratio

            pitch_envelope.release_curve_power = oscillator_config.pitch_envelope_release_curve_power
            pitch_envelope.release_curve = oscillator_config.pitch_envelope_release_curve_ratio

            new_osc.pitch_envelope = pitch_envelope
            new_osc.volume_envelope = volume_envelope

            new_osc.zero_frequency = oscillator_config.zero_frequency
            new_osc.frequency_range = oscillator_config.frequency_range

            new_osc.volume = oscillator_config.volume

            new_osc.initial_phase = oscillator_config.start_phase

            self._oscillators.append(new_osc)

    def render_to_buffer(self, output_buffer: np.ndarray, sample_rate: int, start_sample: int,
                         mode: str = WaveProducerInterface.MODE_REPLACE):

        for pos, oscillator in enumerate(self._oscillators):

            if pos == 0:
                mode = WaveProducerInterface.MODE_REPLACE
            else:
                mode = WaveProducerInterface.MODE_MIX

            oscillator.render_to_buffer(output_buffer, sample_rate, start_sample, mode)

    def get_length_of_full_hit_seconds(self) -> float:
        """
        tells how many samples are needed to fully render a hit with current settings

        It does it by examining volume envelopes of all of the oscillators and
        returning the maximum length of them.
        """
        max_length: float = 0.0
        for osc in self._oscillators:
            current_max = osc.volume_envelope.get_length_seconds()
            if current_max > max_length:
                max_length = current_max

        return max_length



from typing import List

kAmsynthParameter_AmpEnvAttack = 0
kAmsynthParameter_AmpEnvDecay = 1
kAmsynthParameter_AmpEnvSustain = 2
kAmsynthParameter_AmpEnvRelease = 3
kAmsynthParameter_Oscillator1Waveform = 4
kAmsynthParameter_FilterEnvAttack = 5
kAmsynthParameter_FilterEnvDecay = 6
kAmsynthParameter_FilterEnvSustain = 7
kAmsynthParameter_FilterEnvRelease = 8
kAmsynthParameter_FilterResonance = 9
kAmsynthParameter_FilterEnvAmount = 10
kAmsynthParameter_FilterCutoff = 11
kAmsynthParameter_Oscillator2Detune = 12
kAmsynthParameter_Oscillator2Waveform = 13
kAmsynthParameter_MasterVolume = 14
kAmsynthParameter_LFOFreq = 15
kAmsynthParameter_LFOWaveform = 16
kAmsynthParameter_Oscillator2Octave = 17
kAmsynthParameter_OscillatorMix = 18
kAmsynthParameter_LFOToOscillators = 19
kAmsynthParameter_LFOToFilterCutoff = 20
kAmsynthParameter_LFOToAmp = 21
kAmsynthParameter_OscillatorMixRingMod = 22
kAmsynthParameter_Oscillator1Pulsewidth = 23
kAmsynthParameter_Oscillator2Pulsewidth = 24
kAmsynthParameter_ReverbRoomsize = 25
kAmsynthParameter_ReverbDamp = 26
kAmsynthParameter_ReverbWet = 27
kAmsynthParameter_ReverbWidth = 28
kAmsynthParameter_AmpDistortion = 29
kAmsynthParameter_Oscillator2Sync = 30
kAmsynthParameter_PortamentoTime = 31
kAmsynthParameter_KeyboardMode = 32
kAmsynthParameter_Oscillator2Pitch = 33
kAmsynthParameter_FilterType = 34
kAmsynthParameter_FilterSlope = 35
kAmsynthParameter_LFOOscillatorSelect = 36
kAmsynthParameter_FilterKeyTrackAmount = 37
kAmsynthParameter_FilterKeyVelocityAmount = 38
kAmsynthParameter_AmpVelocityAmount = 39
kAmsynthParameter_PortamentoMode = 40
kAmsynthParameterCount = 41


# 	enum class Law {
# 		kLinear,		// offset + base * value
# 		kExponential,	// offset + base ^ value
# 		kPower			// offset + value ^ base
# 	};

class AmsynthParameterLaw:
    EXPONENTIAL = 'exponential'
    POWER = 'power'
    LINEAR = 'linear'


# typedef enum {
# 	KeyboardModePoly,
# 	KeyboardModeMono,
# 	KeyboardModeLegato,
# } KeyboardMode;
#
# typedef enum {
# 	PortamentoModeAlways,
# 	PortamentoModeLegato
# } PortamentoMode;

KeyboardModePoly = 0
KeyboardModeMono = 1
KeyboardModeLegato = 2

PortamentoModeAlways = 0
PortamentoModeLegato = 1

# 	enum class Type {
# 		kLowPass,
# 		kHighPass,
# 		kBandPass,
# 		kBandStop,
# 		kBypass
# 	};
#
# 	enum class Slope {
# 		k12,
# 		k24,
# 	};

kLowPass = 0
kHighPass = 1
kBandPass = 2
kBandStop = 3
kBypass = 4

k12 = 0
k24 = 2


class AmsynthParameter:

    # 					Parameter		(const std::string &name = "unused", Param id = kAmsynthParameterCount,
    # 									 float value = 0.0, float min = 0.0, float max = 1.0, float inc = 0.0,
    # 									 Law law = Law::kLinear, float base = 1.0, float offset = 0.0,
    # 									 const std::string &label = "");

    def __init__(self, name: str = "unused",
                 param_id: int = kAmsynthParameterCount,
                 default: float = 0.0,
                 min_val: float = 0.0,
                 max_val: float = 1.0,
                 inc: float = 0.0,
                 law: str = AmsynthParameterLaw.LINEAR,
                 base: float = 1.0, offset: float = 0.0,
                 label: str = ""):
        self.name: str = name
        self.id: int = param_id
        self.min: float = min_val
        self.max: float = max_val
        self.inc: float = inc
        self.default: float = default
        self.law: str = law
        self.base: float = base
        self.offset: float = offset
        self.label: str = label


class AmsynthTimeParameter(AmsynthParameter):
    def __init__(self, name: str, param_id: int):
        super().__init__(name=name, param_id=param_id, min_val=0.0, max_val=2.5,
                         default=0.0, inc=0.0,
                         law=AmsynthParameterLaw.POWER,
                         base=3, offset=0.0005,
                         label="s")


class AmsynthParameterNormalizer:
    """
    This class converts Amsynth parameter values from bank files into [0..1] ranges to be accepted as vst param values
    """

    def __init__(self):
        self._parameters: List[AmsynthParameter] = [
            AmsynthTimeParameter("amp_attack", kAmsynthParameter_AmpEnvAttack),
            AmsynthTimeParameter("amp_decay", kAmsynthParameter_AmpEnvDecay),
            AmsynthParameter("amp_sustain", kAmsynthParameter_AmpEnvSustain, 1),
            AmsynthTimeParameter("amp_release", kAmsynthParameter_AmpEnvRelease),
            AmsynthParameter("osc1_waveform", kAmsynthParameter_Oscillator1Waveform, 2, 0, 4, 1),
            AmsynthTimeParameter("filter_attack", kAmsynthParameter_FilterEnvAttack),
            AmsynthTimeParameter("filter_decay", kAmsynthParameter_FilterEnvDecay),
            AmsynthParameter("filter_sustain", kAmsynthParameter_FilterEnvSustain, 1),
            AmsynthTimeParameter("filter_release", kAmsynthParameter_FilterEnvRelease),
            AmsynthParameter("filter_resonance", kAmsynthParameter_FilterResonance, 0, 0, 0.97),
            AmsynthParameter("filter_env_amount", kAmsynthParameter_FilterEnvAmount, 0, -16, 16),
            AmsynthParameter("filter_cutoff", kAmsynthParameter_FilterCutoff, 1.5, -0.5, 1.5, 0,
                             AmsynthParameterLaw.EXPONENTIAL, 16, 0),
            AmsynthParameter("osc2_detune", kAmsynthParameter_Oscillator2Detune, 0, -1, 1, 0,
                             AmsynthParameterLaw.EXPONENTIAL, 1.25, 0),
            AmsynthParameter("osc2_waveform", kAmsynthParameter_Oscillator2Waveform, 2, 0, 4, 1),
            AmsynthParameter("master_vol", kAmsynthParameter_MasterVolume, 0.67, 0, 1, 0, AmsynthParameterLaw.POWER, 2,
                             0),
            AmsynthParameter("lfo_freq", kAmsynthParameter_LFOFreq, 0, 0, 7.5, 0, AmsynthParameterLaw.POWER, 2, 0,
                             "Hz"),
            AmsynthParameter("lfo_waveform", kAmsynthParameter_LFOWaveform, 0, 0, 6, 1),
            AmsynthParameter("osc2_range", kAmsynthParameter_Oscillator2Octave, 0, -3, 4, 1,
                             AmsynthParameterLaw.EXPONENTIAL, 2, 0),
            AmsynthParameter("osc_mix", kAmsynthParameter_OscillatorMix, 0, -1, 1),
            AmsynthParameter("freq_mod_amount", kAmsynthParameter_LFOToOscillators, 0, 0, 1.25992105, 0,
                             AmsynthParameterLaw.POWER, 3, -1),
            AmsynthParameter("filter_mod_amount", kAmsynthParameter_LFOToFilterCutoff, -1, -1, 1),
            AmsynthParameter("amp_mod_amount", kAmsynthParameter_LFOToAmp, -1, -1, 1),
            AmsynthParameter("osc_mix_mode", kAmsynthParameter_OscillatorMixRingMod, 0, 0, 1, 0),
            AmsynthParameter("osc1_pulsewidth", kAmsynthParameter_Oscillator1Pulsewidth, 1),
            AmsynthParameter("osc2_pulsewidth", kAmsynthParameter_Oscillator2Pulsewidth, 1),
            AmsynthParameter("reverb_roomsize", kAmsynthParameter_ReverbRoomsize),
            AmsynthParameter("reverb_damp", kAmsynthParameter_ReverbDamp),
            AmsynthParameter("reverb_wet", kAmsynthParameter_ReverbWet),
            AmsynthParameter("reverb_width", kAmsynthParameter_ReverbWidth, 1),
            AmsynthParameter("distortion_crunch", kAmsynthParameter_AmpDistortion, 0, 0, 0.9),
            AmsynthParameter("osc2_sync", kAmsynthParameter_Oscillator2Sync, 0, 0, 1, 1),
            AmsynthParameter("portamento_time", kAmsynthParameter_PortamentoTime, 0.0, 0.0, 1.0),
            AmsynthParameter("keyboard_mode", kAmsynthParameter_KeyboardMode, KeyboardModePoly, 0, KeyboardModeLegato,
                             1),
            AmsynthParameter("osc2_pitch", kAmsynthParameter_Oscillator2Pitch, 0, -12, +12, 1),
            AmsynthParameter("filter_type", kAmsynthParameter_FilterType, kLowPass, kLowPass, kBypass, 1),
            AmsynthParameter("filter_slope", kAmsynthParameter_FilterSlope, k24, k12, k24, 1),
            AmsynthParameter("freq_mod_osc", kAmsynthParameter_LFOOscillatorSelect, 0, 0, 2, 1),
            AmsynthParameter("filter_kbd_track", kAmsynthParameter_FilterKeyTrackAmount, 1),
            AmsynthParameter("filter_vel_sens", kAmsynthParameter_FilterKeyVelocityAmount, 1),
            AmsynthParameter("amp_vel_sens", kAmsynthParameter_AmpVelocityAmount, 1),
            AmsynthParameter("portamento_mode", kAmsynthParameter_PortamentoMode, PortamentoModeAlways)
        ]

    def normalize(self, param_name: str, denormalized_value: float):
        param = self._find_param_by_name(param_name)
        normalized_value = (denormalized_value - param.min) / (param.max - param.min)

        if normalized_value < 0 or normalized_value > 1:
            raise ValueError(
                f'{param_name}: normalized value of {normalized_value} derived from denormalized_value={denormalized_value} is outside [0.0, 1.0] range (error: ce350d7b)')

        return normalized_value

    def denormalize(self, param_name: str, normalized_value: float) -> float:
        if normalized_value < 0 or normalized_value > 1:
            raise ValueError(
                f'{param_name}: normalized value {normalized_value} is outside [0.0, 1.0] range (error: ce350d7b)')

        param = self._find_param_by_name(param_name)
        return param.min + (param.max - param.min) * normalized_value

    def _find_param_by_name(self, param_name: str) -> AmsynthParameter:
        for parameter in self._parameters:
            if parameter.name == param_name:
                return parameter

        raise ValueError(f'unable to find an amsynth parameter named {param_name} (error: 71c9a67c)')



from dataclasses import dataclass
from typing import List

from py_headless_daw.dsp_utils.wave_types import WAVE_TYPE_SINE


@dataclass
class OscillatorConfig:
    volume: float = 1
    zero_frequency: float = 440
    frequency_range: float = 1000
    start_phase: float = 0
    wave_type: str = WAVE_TYPE_SINE

    volume_envelope_attack_time: float = 0.01
    volume_envelope_attack_curve_ratio: float = 0
    volume_envelope_attack_curve_power: float = 1

    volume_envelope_decay_time: float = 0.3
    volume_envelope_decay_curve_ratio: float = 1
    volume_envelope_decay_curve_power: float = 2

    volume_envelope_sustain_level: float = 0
    volume_envelope_sustain_time: float = 0

    volume_envelope_release_time: float = 1
    volume_envelope_release_curve_ratio: float = 0
    volume_envelope_release_curve_power: float = 2

    pitch_envelope_attack_time: float = 0.01
    pitch_envelope_attack_curve_ratio: float = 0
    pitch_envelope_attack_curve_power: float = 2

    pitch_envelope_decay_time: float = 0.4
    pitch_envelope_decay_curve_ratio: float = 0.5
    pitch_envelope_decay_curve_power: float = 2

    pitch_envelope_sustain_level: float = 0
    pitch_envelope_sustain_time: float = 0
    pitch_envelope_release_time: float = 1
    pitch_envelope_release_curve_ratio: float = 1
    pitch_envelope_release_curve_power: float = 2


class DrumSynthGeneratorConfig:
    def __init__(self, oscillators_configs: List[OscillatorConfig]):
        self.oscillator_configs = oscillators_configs



from abc import abstractmethod, ABC
import random


class Event(ABC):

    TYPE_MIDI = 'midi'
    TYPE_TRACK_SAMPLER = 'track_sampler'
    TYPE_PARAMETER_VALUE = 'param_value'

    @property
    @abstractmethod
    def type(self):
        pass

    def __init__(self, sample_position: int):
        self.sample_position: int = sample_position
        self.probability: float = 1.0

    def get_sample_position(self) -> int:
        return self.sample_position

    def should_happen(self) -> bool:
        return random.uniform(0.0, 1.0) < self.probability
